# Chess Analyses by Jan van Reek
# http://www.endgame.nl/index.html

JvR_links = (
    ("http://www.endgame.nl/match.htm", "http://www.endgame.nl/MATCHPGN.ZIP"),
    ("http://www.endgame.nl/bad1870.htm", "http://www.endgame.nl/bad1870.pgn"),
    ("http://www.endgame.nl/wfairs.htm", "http://www.endgame.nl/wfairs.pgn"),
    ("http://www.endgame.nl/russia.html", "http://www.endgame.nl/Russia.pgn"),
    ("http://www.endgame.nl/wien.htm", "http://www.endgame.nl/wien.pgn"),
    ("http://www.endgame.nl/london1883.htm", "http://www.endgame.nl/london.pgn"),
    ("http://www.endgame.nl/neur1896.htm", "http://www.endgame.nl/neur1896.pgn"),
    ("http://www.endgame.nl/newyork.htm", "http://www.endgame.nl/newy.pgn"),
    ("http://www.endgame.nl/seaside.htm", "http://www.endgame.nl/seaside.pgn"),
    ("http://www.endgame.nl/CSpr1904.htm", "http://www.endgame.nl/cs1904.pgn"),
    ("http://www.endgame.nl/stpeter.htm", "http://www.endgame.nl/stp1909.pgn"),
    ("http://www.endgame.nl/stpeter.htm", "http://www.endgame.nl/stp1914.pgn"),
    ("http://www.endgame.nl/berlin1928.htm", "http://www.endgame.nl/berlin.pgn"),
    ("http://www.endgame.nl/bad.htm", "http://www.endgame.nl/bad.pgn"),
    ("http://www.endgame.nl/nimzowitsch.htm", "http://www.endgame.nl/nimzowitsch.pgn"),
    ("http://www.endgame.nl/mostrau.htm", "http://www.endgame.nl/mostrau.pgn"),
    ("http://www.endgame.nl/early.htm", "http://www.endgame.nl/early.pgn"),
    ("http://www.endgame.nl/bled1931.htm", "http://www.endgame.nl/Alekhine.pgn"),
    ("http://www.endgame.nl/nott1936.htm", "http://www.endgame.nl/nott1936.pgn"),
    ("http://www.endgame.nl/wbm.htm", "http://www.endgame.nl/wbm.pgn"),
    ("http://www.endgame.nl/AVRO1938.htm", "http://www.endgame.nl/avro1938.pgn"),
    ("http://www.endgame.nl/salz1942.htm", "http://www.endgame.nl/salz1942.pgn"),
    ("http://www.endgame.nl/itct.html", "http://www.endgame.nl/itct.pgn"),
    ("http://www.endgame.nl/zurich1953.htm", "http://www.endgame.nl/zurich.pgn"),
    ("http://www.endgame.nl/spassky.htm", "http://www.endgame.nl/SPASSKY.ZIP"),
    ("http://www.endgame.nl/dallas1957.htm", "http://www.endgame.nl/dallas57.pgn"),
    ("http://www.endgame.nl/capamem.htm", "http://www.endgame.nl/capamem.pgn"),
    ("http://www.endgame.nl/kortschnoj.htm", "http://www.endgame.nl/korchnoi.pgn"),
    ("http://www.endgame.nl/planinc.htm", "http://www.endgame.nl/Planinc.pgn"),
    ("http://www.endgame.nl/planinc.htm", "http://www.endgame.nl/memorial.pgn"),
    ("http://www.endgame.nl/Piatigorsky.htm", "http://www.endgame.nl/piatigorsky.pgn"),
    ("http://www.endgame.nl/ussr7079.htm", "http://www.endgame.nl/ussr6591.pgn"),
    ("http://www.endgame.nl/tilburg.htm", "http://www.endgame.nl/tilburg.pgn"),
    ("http://www.endgame.nl/dglory.htm", "http://www.endgame.nl/dglory.pgn"),
    ("http://www.endgame.nl/bugojno.htm", "http://www.endgame.nl/Bugojno.pgn"),
    ("http://www.endgame.nl/montreal.htm", "http://www.endgame.nl/mon1979.pgn"),
    ("http://www.endgame.nl/moscow88.htm", "http://www.endgame.nl/ussr88.pgn"),
    ("http://www.endgame.nl/skelleftea.htm", "http://www.endgame.nl/skel1989.pgn"),
    ("http://www.endgame.nl/vsb.htm", "http://www.endgame.nl/vsb.pgn"),
    ("http://www.endgame.nl/dortmund.htm", "http://www.endgame.nl/dortmund.pgn"),
    ("http://www.endgame.nl/Barca.html", "http://www.endgame.nl/Barca.pgn"),
    ("http://www.endgame.nl/Madrid.html", "http://www.endgame.nl/Madrid.pgn"),
    ("http://www.endgame.nl/costa_del_sol.html", "http://www.endgame.nl/Costa.pgn"),
    ("http://www.endgame.nl/Palma.html", "http://www.endgame.nl/Palma.pgn"),
    ("http://www.endgame.nl/olot.html", "http://www.endgame.nl/Olot.pgn"),
    ("http://www.endgame.nl/LasPalmas.html", "http://www.endgame.nl/lpalm96.pgn"),
    ("http://www.endgame.nl/DosH.htm", "http://www.endgame.nl/DosH.pgn"),
    ("http://www.endgame.nl/wijk.htm", "http://www.endgame.nl/corus.pgn"),
    ("http://www.endgame.nl/tal.html", "http://www.endgame.nl/Tal.pgn"),
    ("http://www.endgame.nl/cc.htm", "http://www.endgame.nl/cc.pgn"),
    ("http://www.endgame.nl/sofia.htm", "http://www.endgame.nl/sofia.pgn"),
    ("http://www.endgame.nl/linares.htm", "http://www.endgame.nl/linares.pgn"),
    ("http://www.endgame.nl/Bilbao.html", "http://www.endgame.nl/Bilbao.pgn"),
    ("http://www.endgame.nl/nanjing.html", "http://www.endgame.nl/Nanjing.pgn"),
    ("http://www.endgame.nl/dchamps.htm", "http://www.endgame.nl/dch.pgn"),
    ("http://www.endgame.nl/dsb.htm", "http://www.endgame.nl/dsb.pgn"),
    ("http://www.endgame.nl/cc-history.htm", "http://www.endgame.nl/cc-history.pgn"),
    ("http://www.endgame.nl/hastings.htm", "http://www.endgame.nl/hastings.pgn"),
    ("http://www.endgame.nl/ibm.htm", "http://www.endgame.nl/IBM.pgn"),
    ("http://www.endgame.nl/gambits.htm", "http://www.endgame.nl/gambit.pgn"),
    ("http://www.endgame.nl/trebitsch.htm", "http://www.endgame.nl/Trebitsch.pgn"),
    ("http://www.endgame.nl/cloister.htm", "http://www.endgame.nl/TerApel.pgn"),
    ("http://www.endgame.nl/Biel.html", "http://www.endgame.nl/Biel.pgn"),
    ("http://www.endgame.nl/USA.html", "http://www.endgame.nl/USA.pgn"),
    ("http://www.endgame.nl/uk.html", "http://www.endgame.nl/UK.pgn"),
    ("http://www.endgame.nl/olympiads.html", "http://www.endgame.nl/olympiads.pgn"),
    ("http://www.endgame.nl/lone_pine.html", "http://www.endgame.nl/lonepine.pgn"),
    ("http://www.endgame.nl/staunton.html", "http://www.endgame.nl/Staunton.pgn"),
    ("http://www.endgame.nl/Hoogeveen.html", "http://www.endgame.nl/crown.pgn"),
    ("http://www.endgame.nl/paoli.html", "http://www.endgame.nl/Paoli.pgn"),
    ("http://www.endgame.nl/endgame.htm", "http://www.endgame.nl/endgame.pgn"),
    ("http://www.endgame.nl/estrin.html", "http://www.endgame.nl/Estrin.pgn"),
    ("http://www.endgame.nl/Argentina.html", "http://www.endgame.nl/Argentina.pgn"),
    ("http://www.endgame.nl/comeback.html", "http://www.endgame.nl/comeback.pgn"),
    ("http://www.endgame.nl/strategy.htm", "http://www.endgame.nl/strategy.pgn"),
    ("http://www.endgame.nl/computer.html", "http://www.endgame.nl/computer.pgn"),
    (
        "http://www.endgame.nl/correspondence.html",
        "http://www.endgame.nl/gambitnimzo.pgn",
    ),
    (
        "http://web.inter.nl.net/hcc/rekius/buckle.htm",
        "http://web.inter.nl.net/hcc/rekius/buckle.pgn",
    ),
    (
        "http://web.inter.nl.net/hcc/rekius/euwe.htm",
        "http://web.inter.nl.net/hcc/rekius/euwem.pgn",
    ),
)

JvR = []
for item in JvR_links:
    JvR.append(
        (
            item[0],
            "https://raw.githubusercontent.com/gbtami/JvR-archive/master/%s"
            % item[1][7:],
        )
    )
